from debugot import Debugot, main, RCOORDS, manhattan
from random import randrange
from icfp import BASE, PLAIN, ROBOT, WATER, COORDS, WALL
import sys

# Perpendicular squares
PERP = {
    (0, +1) : ((-1, +1), (+1, +1)),
    (0, -1) : ((-1, -1), (+1, -1)),
    (-1, 0) : ((-1, -1), (-1, +1)),
    (+1, 0) : ((+1, -1), (+1, +1))
    }

FINAL = 0
DEADLOCK_COUNT = 12
DEADLOCK_LIMIT = 5

class StupidClient(Debugot):
    def __init__(self, interactive, timeout):
        Debugot.__init__(self,interactive,timeout)
        self.init()

    def init(self):
        self.mode = self.findPackages
        self.emptyBases = {}            # Bases we know are empty
        self.target = None              # Current target coordinates
        self.path = None                # Current path list to get to target
        self.pathOffset = 0             # offset into self.path
        self.gotThere = None            # Function to call back when we're there
        self.bid = 1                    # Default bid
        self.wanderCount = 0            # How long we wander randomly
        self.visited = []               # The last 10 squares we've been at
        self.deadlockLevel = 0

    def ourFirstTime(self):
        self.addLog("I live!")
        # If we find something within that length, we go there no matter what
        self.nearLength = 2 + (len(self.map) + len(self.map[0])) / 10

    def checkDeadlock(self):
        d = {}
        for x in self.visited:
            d[x] = d.get(x,0) + 1
            if d[x] > DEADLOCK_LIMIT:
                self.visited = []
                self.addLog('Deadlock for: %d,%d' % x)
                return 1
        self.deadlockLevel = max(d.values())
        return 0
        

    def think(self):
        try:
            self.visited.append(self.getPosition())
            self.visited = self.visited[-DEADLOCK_COUNT:]
            if self.checkDeadlock():
                return self.wanderFor(5)
            self.recLevel = 0
            return self.mode()
        except:
            if not FINAL: raise

            # Reset everything!
            self.addLog('Exception! init 6!')
            self.init()

            return 'Rebooting', '1 pick'
            

    def getNearMap(self,size=2):
        "Generate a map for the nearest few squares."
        cx,cy = self.getPosition()
        d = {}
        for x in xrange(-size,+size+1):
            for y in xrange(-size,+size+1):
                ax = x+cx; ay = y+cy
                if ax < 0 or ay < 0 or ax > (self.boardX+2) or ay > (self.boardY+2):
                    d[x,y] = WALL
                else:
                    d[x,y] = self.map[ay][ax]

        for r in self.knownRobots.values():
            x = r.x - cx; y = r.y - cy
            if x < -size or x > +size or y < -size or y > +size:
                continue
            d[x,y] = ROBOT

        return d

    def robotCommand(self, name, args):
        if name == 'e':
            return self.evade(force=1)
        else:
            assert 0

    def wander(self):
        "Can't find any bases?"
        self.wanderCount -= 1
        if self.wanderCount <= 0:
            return self.setMode(self.findPackages)


        return 'Evading', self.evade(force=1)

    def evade(self, force=0):
        "Move to a non-dangerous spot. Quickly if there are bots a-coming"
        # Find a nice safe place
        score = {}
        d = self.getNearMap()
        robots = 0
        for dx, dy in COORDS.values():
            s = 0
            if d[dx,dy] in (BASE, PLAIN):
                s += 2
                for dx2, dy2 in COORDS.values():
                    if not (dx+dx2 or dy+dy2):
                        continue
                    # We don't like being near water
                    c = d[dx+dx2,dy+dy2]
                    if c == WATER:
                        s -= 2
                    elif c == ROBOT:
                        s -= 1
                    elif c in (BASE, PLAIN):
                        s += 1
            elif d[dx,dy] in (WALL, WATER):
                s = -9
            elif d[dx,dy] == ROBOT:
                s = -1
            score[dx,dy] = s
        l = score.items()
        l.sort(lambda x,y: -cmp(x[1], y[1]))
        #self.addLog("%s" % ';'.join(['%+d,%+d:%+d' % (x,y,v) for (x,y), v in score.items()]))
        if l:
            if force or l[0][1] > 0:
                # good way to go
                if l[0][1] < -5: # really really bad
                    return '1 pick'
                else:
                    return '%d Move %s' % ((1 + robots*3), RCOORDS[l[0][0]])
        self.addLog('evade gives up')
        return None

    def followPath(self):
        "just follow the path"
        if self.recLevel > 5:
            return 'Recursing??', '1 pick'
        self.recLevel += 1
        
        cur = self.getPosition()
        if cur == self.target:
            return self.setMode(self.gotThere)
        elif self.pathOffset >= len(self.path):
            # What happened? Maybe we got pushed
            self.addLog('At the end of the path but something went wrong. Realigning')
            return self.retarget()

        # Going to deliver, but I've lost all my packages!
        if self.gotThere == self.deliver and not self.myPackages:
            self.addLog("Oh dear, I've lost my packages!")
            return self.setMode(self.findPackages)

        next = self.path[self.pathOffset]
            
        cur = self.getPosition()
        dx = next[1] - cur[0]           # not reversal of order, ugh
        dy = next[0] - cur[1]
        dir = RCOORDS.get((dx,dy))
        if dir is None:
            self.addLog('Off-path?  %d,%d -> %d,%d' % (
                cur[0], cur[1], next[1], next[0]))
            return self.retarget()

        nodes = self.path[self.pathOffset:]
        if len(nodes) > 10:
            nodes = nodes[:10]
            extra = "..."
        else:
            extra = ""

        r = "Go to (%d,%d): %s%s then %s\n " % (
            self.target[0], self.target[1],
            ' '.join(['%d,%d' % (x,y) for x,y in nodes]),
            extra,
            self.gotThere.name)
        
        self.pathOffset += 1
        self.bid = max(1, self.bid - 1)
        return r + 'Moving', '%d move %s' % (self.bid, dir)

    def retarget(self):
        self.bid += 3
        if self.bid > 12:
            self.addLog("Help! I'm being repressed!")
            self.bid = 1
            return self.wanderFor(5)

        # Check if our current position is still somewhere on our path
        pos = self.getPosition()
        try:
            offset = self.path.index((pos[1],pos[0]))
            self.pathOffset = offset + 1
            self.addLog('Skipping to element %d in path' % offset)
            return self.followPath()
        except ValueError:
            return self.gotoXY(self.target[0], self.target[1], self.gotThere)
        

    def goto(self, path, fun):
        "Start moving toward a path. First element in that path should be current pos"
        self.target = path[-1][1], path[-1][0]
        self.path = path[1:]
        self.pathOffset = 0
            
        self.gotThere = fun
        self.addLog('Destination: %d,%d (%d steps away)' % (self.target[0], self.target[1],
                                                            len(self.path)))
        if len(self.path) == 0:
            return self.setMode(fun)
        else:
            return self.setMode(self.followPath)

    def gotoXY(self, x, y, fun):
        "Combines findPath + goto"
        path = self.findpath((x,y))
        if not path:
            # Uh oh. What now?
            self.addLog('%d,%d is UNREACHABLE!' % (x,y))
            return self.setMode(self.findPackages)
        else:
            return self.goto(path, fun)

    def setMode(self, mode):
        "Set a mode and execute it right away"
        self.mode = mode
        return self.mode()

    def beginDelivery(self):
        "Verify that we've picked up stuff"
        if not self.myPackages:
            self.addLog('No packages, going to fill up again')
            return self.setMode(self.findPackages)
        else:
            # Destination... adventure!
            return self.gotoXY(self.myPackages[0].x, self.myPackages[0].y, self.deliver)

    def calculateDangerousBid(self):
        # If there are robots around us, hurry up
        robots = 0
        for x in self.knownRobots.values():
            if x.id != self.number:
                if manhattan((x.x,x.y), self.getPosition()) < 2:
                    robots += 1
        return 1 + robots*3

    def deliver(self):
        "Deliver packages here"
        pos = self.getPosition()
        l = []
        for p in self.myPackages:
            if (p.x,p.y) == pos:
                l.append(p.id)

        if not l:
            # How strange. No packages to be dropped here? Maybe we lost them underway.
            self.addLog('No droppable packages?')
            return self.setMode(self.beginDelivery)

        bid = self.calculateDangerousBid()
        self.addLog("Delivering  packages: bid %d" % bid)
        return 'Delivering the load @%d' % bid, \
               '%d drop %s' % (bid, ' '.join(map(str, l)))
    
        
    deliver.name = 'deliver packages'

    def die(self):
        return 'Goodbye cruel world', '%d pick' % (1+self.money)

    def wanderFor(self, count):
        self.wanderCount = count
        self.addLog('Wandering randomly for %d turns' % count)
        return self.setMode(self.wander)

    def tryPickup(self):
        "We're at somewhere with packages, pick them up"
        pos = self.getPosition()
        plist = self.knownPackages.get(pos, [])
        if not plist:
            # No packages? Maybe we missed the base or it really is empty
            # Keep going
            self.addLog('No packages at %d,%d -- did I miss it?' % pos)
            self.emptyBases[pos] = 1
            return self.setMode(self.findPackages)

        packages = {}; totalWeights = {}
        for p in plist:
            packages.setdefault((p.x, p.y), []).append(p)
            totalWeights[p.x,p.y] = totalWeights.get((p.x,p.y),0) + p.weight

        # For each X,Y coordinates, calculate the reward for getting there
        distances = {}
        for coord, p in packages.items():
            path = self.findpath(coord)
            if path:
                distances[coord] = randrange(0,5) + len(path)
                # Penalty for insufficient load
                if totalWeights[coord] < self.weight:
                    distances[coord] *= 2

        distances = distances.items()
        distances.sort(lambda x,y: cmp(x[1], y[1]))

        if not distances:
            # NO packages are deliverable?
            self.addLog('No reachable package targets? Marking %d as empty' % pos)
            self.emptyBases[pos] = 1
            return self.setMode(self.findPackages)

        # Pick all items (should really try to get the perfect fit)
        target = distances[0][0]

        self.mode = self.beginDelivery
        bid = self.calculateDangerousBid()
        self.addLog("Picking up packages: bid %d" % bid)
        plist = packages[target]
        plist.sort(lambda x,y: cmp(x.weight,y.weight))
        return 'Gonna get me some packages @%d' % bid,  '%d pick %s' % (bid,
                                                              ' '.join([str(x.id) for x in plist]))
    

    tryPickup.name = 'pickup stuff'

    def findPackages(self):
        # Find our closest base or place with packages from here

        targets = []
        tlist = self.bases[:] + self.knownPackages.keys()

        # Sort the order we findpath in based on trivial distance calculation
        # Fuzz it a bit!
        mx, my = self.getPosition()
        coord = (my, mx)
        tlist.sort(lambda x,y: cmp(manhattan(coord, x)/3, manhattan(coord,y)/3))
       
        for coord in tlist:
            if coord in self.emptyBases: continue
            path = self.findpath(coord)
            if path:            # reachable?
                targets.append(path)
                # Don't search for every base if this is fairly nearby
                if len(path) < self.nearLength:
                    self.addLog('<%d, good enough for me' % self.nearLength)
                    break

        # Find shortest distance
        targets.sort(lambda x,y: cmp(len(x), len(y)))
        
        if targets:
            target = targets[0]
            self.addLog('Checked %d targets, going for %d,%d' % ((len(targets),) + target[-1]))

            return self.goto(target, self.tryPickup)

        self.addLog('No bases available')
        return self.wanderFor(10)

if __name__ == '__main__':
    main(StupidClient)

